/*
 * Template output.
 *
 * Copyright (C) 2019, 2020 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#pragma once

/* Filename definitions. */

#define client_filename_c                 "%s/%s_client.c"
#define client_filename_cpp               "%s/%s_client.cc"
#define client_header_filename            "%s/%s_client.h"
#define compound_interfaces_filename      "%s/%s_interfaces.h"
#define interface_filename                "%s/%s_interface.h"
#define server_filename_c                 "%s/%s_server.c"
#define server_filename_cpp               "%s/%s_server.cc"
#define server_header_filename            "%s/%s_server.h"

/* Client templates. */

#define client_prologue \
"#include \"%s_client.h\"\n" \
"#include <ipc/message.h>\n"

#define client_function_body_prologue \
"  ipc_message_t msg;\n" \
"  long err;\n"

#define client_function_body_endpoint_conversion_c \
"  %s _endp = _self.cap;\n"

#define client_function_body_call \
"\n  ipc_message_request(&msg, %s, %s);\n\n" \
"  err = l4_error(msg.tag);\n" \
"  if (err)\n" \
"    return err;\n"

/* Client interface definitions for C++. */

#define client_interface_prologue_cpp \
" : public %s"

#define client_interface_endpoint_declaration_cpp \
"\nprotected:\n" \
"  %s _endp;\n"

#define client_interface_constructor_cpp \
"\n  %s(%s endp) : _endp(endp) { }\n"

/* Client interface definitions for C. */

#define client_interface_declaration_c \
"\nextern iface_%s client_iface_%s;"

#define client_interface_prologue_c \
"\niface_%s client_iface_%s = {"

#define client_interface_member_c \
"\n  .%s = %s,"

#define client_interface_epilogue_c \
"\n};\n"

/* Compound interface dispatcher templates. */

#define compound_dispatch_prologue \
"#include \"%s_server.h\"\n"

#define compound_dispatch_function_prologue \
"  switch (l4_msgtag_label(msg->tag))\n" \
"  {\n"

#define compound_dispatch_epilogue \
"    default:\n" \
"      ipc_message_send_error(msg, -L4_EBADPROTO);\n" \
"      ipc_message_reply(msg);\n" \
"      break;\n" \
"  }\n"

#define compound_dispatch_header_prologue \
"#pragma once\n\n" \
"#include <ipc/message.h>\n" \
"#include \"%s_interface.h\"\n"

/* Compound interface class definitions. */

#define compound_interface_prologue_cpp \
"#pragma once\n\n" \
"#include \"%s_interfaces.h\"\n\n" \
"#ifdef __cplusplus\n\n" \
"class %s :"

#define compound_interface_prologue_c \
"#pragma once\n\n" \
"#include \"%s_interfaces.h\"\n\n" \
"typedef struct\n" \
"{\n"

#define compound_interface_epilogue_cpp \
"\n{\n};\n\n" \
"#endif /* __cplusplus */\n"

#define compound_interface_epilogue_c \
"\n} iface_%s;\n"

#define compound_interfaces_prologue \
"#pragma once\n\n"

/* Convert between object types by copying the reference type value into the
   temporarily allocated structure member which should appear first in the
   reference type. */

#define compound_interface_conversion \
"\n#define convert_to_%s(_self) (memcpy(&((%s) {.iface=_self->iface->to_%s}).ref, &_self->ref, sizeof(ref_%s)))\n"

/* Dispatch functions. */

#define dispatch_function_signature \
"\nvoid dispatch_%s(ipc_message_t *msg, %s *_self)"

/* Message handling functions. */

#define handle_function_signature \
"\nvoid handle_%s(ipc_message_t *msg, %s *_self)"

#define handle_function \
"  ipc_message_open(msg);\n" \
"  dispatch_%s(msg, _self);\n" \
"  ipc_message_discard(msg);\n"

/* General header template. */

#define header_prologue \
"#pragma once\n\n" \
"#include <l4/sys/err.h>\n" \
"#include <l4/sys/types.h>\n"

/* Server templates. */

#define server_header_prologue \
"#pragma once\n\n" \
"#include <ipc/message.h>\n"

#define server_prologue \
"#include \"%s_server.h\"\n"

/* Server wrapper function templates. */

#define server_function_signature \
"\nlong %s_%s(ipc_message_t *msg, %s *_self)"

#define server_function_body_prologue \
"  long err;\n"

#define server_function_body_unused_message \
"\n  (void) msg;\n"

#define server_function_body_call \
"  if (err)\n" \
"    return err;\n"

#define server_function_body_epilogue \
"\n  return L4_EOK;\n"

#define server_completion_function_signature_prologue \
"\nlong complete_%s(l4_cap_idx_t _endp"

#define server_completion_function_body_prologue \
"  ipc_message_t msg;\n"

#define server_completion_function_body_epilogue \
"\n  ipc_message_send(&msg, %s, %s);\n\n" \
  "  return l4_error(msg.tag);\n"

#define server_initiation_function_body_epilogue \
"\n  ipc_message_prepare(msg);\n\n" \
  "  return L4_EOK;\n"

/* Server dispatch function templates. */

#define server_function_dispatcher_body_epilogue \
"    default:\n" \
"      ipc_message_send_error(msg, -L4_EBADPROTO);\n" \
"      ipc_message_reply(msg);\n" \
"      break;\n" \
"  }\n"

/* Interface class definitions. */

#define interface_prologue_cpp \
"\n#ifdef __cplusplus\n" \
"\nclass %s"

#define interface_prologue_c \
"\ntypedef struct" \

#define interface_body_begin \
"\n{"

#define interface_signatures_prologue_cpp \
"\npublic:"

#define interface_epilogue_cpp \
"};\n\n" \
"#endif /* __cplusplus */\n"

#define interface_epilogue_c \
"\n} iface_%s;\n"

#define interface_function_signature_prologue \
"\n%slong %s("

#define interface_struct_member_function_signature_prologue \
"\n  long (*%s)("

#define interface_unimplemented_method_prologue_cpp \
"\n  {"

#define interface_unimplemented_method_epilogue_cpp \
" return -L4_EBADPROTO; }\n"

#define interface_unused_parameter \
" (void) %s;"

/* Reference and object type definitions for interfaces. */

#define ref_type_definition_c \
"\ntypedef union { %s cap; void *ptr; } ref_%s;\n"

/* Note that the conversion operation relies on the ref member appearing first.
   See: compound_interface_conversion */

#define object_type_definition_c \
"\ntypedef struct\n" \
"{\n" \
"  ref_%s ref;\n" \
"  iface_%s *iface;\n" \
"\n} %s;\n"

#define expected_items_definition \
"\n#define %s_expected_items %d\n"

/* Opcode definitions. */

#define opcode_enumeration_prologue \
"\nenum opcodes_%s\n{\n"

#define opcode_identifier \
"  opcode_%s_%s"

#define opcode_enumeration_epilogue \
"};\n"

/* Message structure definitions. */

#define structure_prologue \
"\nstruct %s_words_%s\n{\n"

#define structure_opcode_member \
"  %s _op;\n"

#define structure_epilogue \
"};\n"

/* Message access templates. */

#define message_accessor_declaration \
"  struct %s_words_%s *%s_words;\n"

#define message_accessor_initialisation \
"\n  %s_words = (struct %s_words_%s *)"

#define message_accessor_reading_initialiser \
" ipc_message_get_word_address(%smsg, 0);\n"

#define message_accessor_writing_initialiser \
" ipc_message_reserve_words(%smsg, sizeof(struct %s_words_%s));\n"

/* Dispatch templates. */

#define dispatch_function_interface_case \
"    case %s:\n" \
"      dispatch_%s(msg, %s);\n" \
"      break;\n\n"

#define dispatch_function_wrapper_case \
"    case %s:\n" \
"      ipc_message_send_error(msg, %s_%s(msg, %s));\n" \
"      break;\n\n"

#define dispatch_function_reply_wrapper_case \
"    case %s:\n" \
"      ipc_message_send_error(msg, %s_%s(msg, %s));\n" \
"      ipc_message_reply(msg);\n" \
"      break;\n\n"

/* Tokens. */

#define BEGIN_FUNCTION "\n{\n"
#define END_FUNCTION "}\n"

#define COMPLETE_MEMBER ";\n"
#define INDENT_MEMBER "  "

#define COMPLETE_SIGNATURE ";\n"
#define END_SIGNATURE "\n"

/* Texts. */

#define help_text \
"Usage: %s [ <options> ] <filename> ...\n\n" \
"Generate source code for the interfaces in the supplied files.\n\n" \
"The following options are supported:\n\n" \
"--all or -a          Produce all kinds of output for a given language.\n\n" \
"--client or -c       Generate client code.\n\n" \
"--comp or -C         Generate a compound/composite/component interface using\n" \
"                     the given filename prefix.\n\n" \
"--dir or -d          Create output in the indicated directory.\n\n" \
"--headers or -h      Produce header files.\n\n" \
"--help or -?         Show this message.\n\n" \
"--interfaces or -i   Produce interface header files.\n\n" \
"--language or -l     Generate code in the indicated language.\n\n" \
"--comp-name or -N    Indicate the compound/composite/component interface name.\n\n" \
"--routines or -r     Produce program routines.\n\n" \
"--server or -s       Generate server code.\n\n" \
"--verbose or -v      Show verbose output reporting the processed interfaces.\n\n" \
"--version or -V      Show version information.\n\n"
